import 'dart:convert';
import 'dart:io';

import 'package:html/dom.dart';
import 'package:html/parser.dart';

// TODO: Put public facing types in this file.

class SnapView {
  String domain;
  String author;
  String title;
  String summary;
  String picture;
  String videoUri;
  SnapView(
      {this.domain = '',
      this.author = '',
      required this.title,
      required this.summary,
      this.picture = '',
      this.videoUri = ''});

  SnapView.emptyView()
      : domain = '',
        author = 'Unknown',
        title = 'Unknown',
        summary = 'N/A',
        picture = '',
        videoUri = '';
}

Future<String?> fetchPage(String uri) async {
  final client = HttpClient();
  final resq = await client.getUrl(Uri.parse(uri));
  final resp = await resq.close();
  if (resp.statusCode >= 200 && resp.statusCode < 300) {
    final bytes = (await resp.toList()).expand((element) => element).toList();
    return utf8.decode(bytes);
  }
}

Future<SnapView> analyzePage(String htmlStr) async {
  final document = parse(htmlStr);
  final view = SnapView.emptyView();
  processOgc(document, view);
  processTitle(document, view);
  processSummary(document, view);

  final contentElement = findContentBody(document.body!.children);
  if (contentElement != null) {
    if (view.summary == 'N/A') {
      view.summary = contentElement.text.replaceAll(' ', '').substring(0, 144);
    }
    final img = findImg(contentElement);
    if (img != null) {
      if (view.picture == '') {
        view.picture = img.attributes['src']!;
      }
    }
  }
  processFavicon(document, view);
  return view;
}

Element? findImg(Element tar) {
  if (tar.children != null) {
    for (var ele in tar.children) {
      if (ele.localName == 'img' && ele.attributes['src']!.startsWith('http')) {
        return ele;
      } else {
        if (ele.children != null) {
          final rs = findImg(ele);
          if (rs != null) {
            return rs;
          }
        }
      }
    }
  }
}

Element? findContentBody(List<Element> childs) {
  final matched = childs
      .map((element) {
        if (element.localName == 'article' ||
            element.classes
                .where((classname) =>
                    classname.contains('content') ||
                    classname.contains('article'))
                .toList()
                .isNotEmpty) {
          return element;
        }
      })
      .where((element) => element != null)
      .toList();
  if (matched.isNotEmpty) {
    return matched.first;
  } else {
    for (var ele in childs) {
      final rs = findContentBody(ele.children);
      if (rs != null) {
        return rs;
      }
    }
  }
}

void processFavicon(Document document, SnapView view) {
  if (view.picture == '') {
    document.head!.children.forEach((element) {
      if (element.localName == 'link' && element.attributes['rel'] == 'icon') {
        view.picture = element.attributes['href']!;
      }
    });
  }
}

void processOgc(Document document, SnapView view) {
  document.head!.children.forEach((element) {
    if (element.localName == 'meta') {
      final property = element.attributes['property'];
      final value = element.attributes['content'];
      switch (property) {
        case 'og:title':
          view.title = value ?? 'Unknown';
          break;
        case 'og:image':
          view.picture = value ?? '';
          break;
        case 'og:description':
          view.summary = value ?? 'N/A';
          break;
        default:
          break;
      }
    }
  });
}

void processSummary(Document document, SnapView view) {
  if (view.summary != 'N/A') {
    return;
  }
  if (document.head != null) {
    final metas = document.head!.children
        .where((element) =>
            element.localName == 'meta' &&
            element.attributes['name'] == 'description')
        .toList();
    if (metas.isNotEmpty) {
      view.summary = metas.first.attributes['content']!;
    }
  }
}

void processTitle(Document document, SnapView view) {
  if (view.title != 'Unknown') {
    return;
  }
  if (document.head != null) {
    final titles = document.head!.children
        .where((element) => element.localName == 'title')
        .toList();
    if (titles.isNotEmpty) {
      view.title = titles.first.innerHtml;
    }
  }
}

class Snapshot {
  String url;
  Snapshot(this.url);

  Future<SnapView> process() async {
    final htmlStr = await fetchPage(url);
    if (htmlStr != null) {
      return analyzePage(htmlStr);
    } else {
      return SnapView.emptyView();
    }
  }
}
